package gregapi.item.multiitem;

import static gregapi.data.CS.*;
import gregapi.code.ArrayListNoNulls;
import gregapi.code.TagData;
import gregapi.data.CS.ModIDs;
import gregapi.data.TD;
import gregapi.item.IItemEnergy;
import gregapi.item.ItemBase;
import gregapi.item.multiitem.behaviors.IBehavior;
import gregapi.lang.LanguageHandler;
import gregapi.recipes.GT_ModHandler;
import gregapi.util.UT;
import ic2.api.item.ElectricItem;
import ic2.api.item.IElectricItem;
import ic2.api.item.IElectricItemManager;
import ic2.api.item.ISpecialElectricItem;

import java.util.*;

import net.minecraft.dispenser.IBlockSource;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.Container;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.EnumChatFormatting;
import net.minecraft.world.World;
import net.minecraftforge.fluids.FluidStack;
import cpw.mods.fml.common.Optional;

/**
 * @author Gregorius Techneticies
 * 
 * For Custom Items.
 */
@Optional.InterfaceList(value = {
  @Optional.Interface(iface = "ic2.api.item.ISpecialElectricItem", modid = ModIDs.IC2)
, @Optional.Interface(iface = "ic2.api.item.IElectricItemManager", modid = ModIDs.IC2)
})
public abstract class MultiItem extends ItemBase implements ISpecialElectricItem, IElectricItemManager, IItemEnergy {
	private final HashMap<Short, ArrayList<IBehavior<MultiItem>>> mItemBehaviors = new HashMap<Short, ArrayList<IBehavior<MultiItem>>>();
	
	public final HashMap<Short, Long[]> mFluidContainerStats = new HashMap<Short, Long[]>();
	
	/**
	 * Creates the Item using these Parameters.
	 * @param aUnlocalized The Unlocalized Name of this Item. DO NOT START YOUR UNLOCALISED NAME WITH "gt."!!!
	 * @param aGeneratedPrefixList The OreDict Prefixes you want to have generated.
	 */
	public MultiItem(String aModID, String aUnlocalized) {
		super(aModID, aUnlocalized, "Generated Item", null);
        setHasSubtypes(T);
        setMaxDamage(0);
	}
	
	/**
	 * Adds a special Item Behaviour to the Item.
	 * 
	 * Note: the boolean Behaviours sometimes won't be executed if another boolean Behaviour returned true before.
	 * 
	 * @param aMetaValue the Meta Value of the Item you want to add it to. [0 - 32765]
	 * @param aBehavior the Click Behavior you want to add.
	 * @return the Item itself for convenience in constructing.
	 */
	public MultiItem addItemBehavior(int aMetaValue, IBehavior<MultiItem> aBehavior) {
		if (aMetaValue < 0 || aMetaValue >= 32766 || aBehavior == null) return this;
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)aMetaValue);
		if (tList == null) {
			tList = new ArrayListNoNulls<IBehavior<MultiItem>>();
			mItemBehaviors.put((short)aMetaValue, tList);
		}
		tList.add(aBehavior);
		return this;
	}
	
	@Override
	public boolean hasProjectile(TagData aProjectileType, ItemStack aStack) {
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) if (tBehavior.hasProjectile(this, aProjectileType, aStack)) return T;
    	return super.hasProjectile(aProjectileType, aStack);
	}
	
	@Override
	public EntityProjectile getProjectile(TagData aProjectileType, ItemStack aStack, World aWorld, double aX, double aY, double aZ) {
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) {
			EntityProjectile rArrow = tBehavior.getProjectile(this, aProjectileType, aStack, aWorld, aX, aY, aZ);
			if (rArrow != null) return rArrow;
		}
    	return super.getProjectile(aProjectileType, aStack, aWorld, aX, aY, aZ);
	}
	
	@Override
	public EntityProjectile getProjectile(TagData aProjectileType, ItemStack aStack, World aWorld, EntityLivingBase aEntity, float aSpeed) {
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) {
			EntityProjectile rArrow = tBehavior.getProjectile(this, aProjectileType, aStack, aWorld, aEntity, aSpeed);
			if (rArrow != null) return rArrow;
		}
    	return super.getProjectile(aProjectileType, aStack, aWorld, aEntity, aSpeed);
	}
	
    @Override
	public ItemStack onDispense(IBlockSource aSource, ItemStack aStack) {
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) if (tBehavior.canDispense(this, aSource, aStack)) return tBehavior.onDispense(this, aSource, aStack);
    	return super.onDispense(aSource, aStack);
    }
    
    @Override
	public boolean isItemStackUsable(ItemStack aStack) {
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) if (!tBehavior.isItemStackUsable(this, aStack)) return F;
    	return super.isItemStackUsable(aStack);
    }
    
    @Override
    public boolean itemInteractionForEntity(ItemStack aStack, EntityPlayer aPlayer, EntityLivingBase aEntity) {
    	use(aStack, 0, aPlayer);
		isItemStackUsable(aStack);
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) try {
			if (tBehavior.onRightClickEntity(this, aStack, aPlayer, aEntity)) {
				if (aStack.stackSize <= 0) aPlayer.destroyCurrentEquippedItem();
				return T;
			}
			if (aStack.stackSize <= 0) {
				aPlayer.destroyCurrentEquippedItem();
				return F;
			}
		} catch(Throwable e) {
			if (D1) e.printStackTrace(ERR);
		}
    	return F;
    }
    
    @Override
    public boolean onLeftClickEntity(ItemStack aStack, EntityPlayer aPlayer, Entity aEntity) {
    	use(aStack, 0, aPlayer);
		isItemStackUsable(aStack);
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) try {
			if (tBehavior.onLeftClickEntity(this, aStack, aPlayer, aEntity)) {
				if (aStack.stackSize <= 0) aPlayer.destroyCurrentEquippedItem();
				return T;
			}
			if (aStack.stackSize <= 0) {
				aPlayer.destroyCurrentEquippedItem();
				return F;
			}
		} catch(Throwable e) {
			if (D1) e.printStackTrace(ERR);
		}
    	return F;
    }
    
	@Override
	public boolean onItemUse(ItemStack aStack, EntityPlayer aPlayer, World aWorld, int aX, int aY, int aZ, int aSide, float hitX, float hitY, float hitZ) {
		use(aStack, 0, aPlayer);
		isItemStackUsable(aStack);
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) try {
			if (tBehavior.onItemUse(this, aStack, aPlayer, aWorld, aX, aY, aZ, UT.Code.side(aSide), hitX, hitY, hitZ)) {
				if (aStack.stackSize <= 0) aPlayer.destroyCurrentEquippedItem();
				return T;
			}
			if (aStack.stackSize <= 0) {
				aPlayer.destroyCurrentEquippedItem();
				return F;
			}
		} catch(Throwable e) {
			if (D1) e.printStackTrace(ERR);
		}
    	return F;
	}
	
	@Override
	public boolean onItemUseFirst(ItemStack aStack, EntityPlayer aPlayer, World aWorld, int aX, int aY, int aZ, int aSide, float hitX, float hitY, float hitZ) {
		use(aStack, 0, aPlayer);
		isItemStackUsable(aStack);
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) try {
			if (tBehavior.onItemUseFirst(this, aStack, aPlayer, aWorld, aX, aY, aZ, UT.Code.side(aSide), hitX, hitY, hitZ)) {
				if (aStack.stackSize <= 0) aPlayer.destroyCurrentEquippedItem();
				return T;
			}
			if (aStack.stackSize <= 0) {
				aPlayer.destroyCurrentEquippedItem();
				return F;
			}
		} catch(Throwable e) {
			if (D1) e.printStackTrace(ERR);
		}
		return F;
	}
	
    @Override
	public ItemStack onItemRightClick(ItemStack aStack, World aWorld, EntityPlayer aPlayer) {
    	use(aStack, 0, aPlayer);
		isItemStackUsable(aStack);
    	ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) try {
			aStack = tBehavior.onItemRightClick(this, aStack, aWorld, aPlayer);
		} catch(Throwable e) {
			if (D1) e.printStackTrace(ERR);
		}
		return aStack;
    }
    
	@Override
    public final void addInformation(ItemStack aStack, EntityPlayer aPlayer, List aList, boolean aF3_H) {
		String tKey = getUnlocalizedName(aStack) + ".tooltip", tString = LanguageHandler.translate(tKey, tKey);
		if (UT.Code.stringValid(tString) && !tKey.equals(tString)) aList.add(tString);
		
		IItemEnergy tEnergyStats = getEnergyStats(aStack);
		if (tEnergyStats != null) for (TagData tEnergyType : tEnergyStats.getEnergyTypes(aStack)) {
			aList.add(tEnergyType.mChatFormat + "" + tEnergyStats.getEnergyStored(tEnergyType, aStack) + " / " + tEnergyStats.getEnergyCapacity(tEnergyType, aStack) + " "+tEnergyType.getLocalisedNameShort()+" - Size: " + tEnergyStats.getEnergySizeInputRecommended(tEnergyType, aStack) + EnumChatFormatting.GRAY);
		}
		
		Long[] tStats = getFluidContainerStats(aStack);
		if (tStats != null && tStats[0] > 0) {
			FluidStack tFluid = getFluidContent(aStack);
			aList.add(EnumChatFormatting.BLUE + ((tFluid==null?"No Fluids Contained":UT.Fluids.name(tFluid, T))) + EnumChatFormatting.GRAY);
			aList.add(EnumChatFormatting.BLUE + ((tFluid==null?0:tFluid.amount) + "L / " + tStats[0] + "L") + EnumChatFormatting.GRAY);
		}
		
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) aList = tBehavior.getAdditionalToolTips(this, aList, aStack);
		
		addAdditionalToolTips(aList, aStack, aF3_H);
	}
	
	@Override
	public void onUpdate(ItemStack aStack, World aWorld, Entity aPlayer, int aTimer, boolean aIsInHand) {
		ArrayList<IBehavior<MultiItem>> tList = mItemBehaviors.get((short)getDamage(aStack));
		if (tList != null) for (IBehavior<MultiItem> tBehavior : tList) tBehavior.onUpdate(this, aStack, aWorld, aPlayer, aTimer, aIsInHand);
	}
	
	@Override
	public boolean canProvideEnergy(ItemStack aStack) {
		return T;
	}
	
	@Override
	public double getMaxCharge(ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergyCapacity(TD.Energy.EU, aStack);
	}
	
	@Override
	public double getTransferLimit(ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergySizeInputRecommended(TD.Energy.EU, aStack);
	}
	
	@Override
	public double charge(ItemStack aStack, double aCharge, int aTier, boolean aIgnoreTransferLimit, boolean aSimulate) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null || aCharge < V[aTier]) return 0;
		return V[aTier] * tStats.doEnergyInjection(TD.Energy.EU, aStack, V[aTier], (long)(aCharge / V[aTier]), null, null, 0, 0, 0, !aSimulate);
	}
	
	@Override
	public double discharge(ItemStack aStack, double aCharge, int aTier, boolean aIgnoreTransferLimit, boolean aBatteryAlike, boolean aSimulate) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null || aCharge < V[aTier]) return 0;
		return V[aTier] * tStats.doEnergyExtraction(TD.Energy.EU, aStack, V[aTier], (long)(aCharge / V[aTier]), null, null, 0, 0, 0, !aSimulate);
	}
	
	@Override
	public double getCharge(ItemStack aStack) {
		return getRealCharge(aStack);
	}
	
	@Override
	public boolean canUse(ItemStack aStack, double aAmount) {
		return F;
	}
	
	@Override
	public boolean use(ItemStack aStack, double aAmount, EntityLivingBase aPlayer) {
		return F;
	}
	
	@Override
	public void chargeFromArmor(ItemStack aStack, EntityLivingBase aPlayer) {
		if (aPlayer == null || aPlayer.worldObj.isRemote) return;
		for (int i = 1; i < 5; i++) {
			ItemStack tArmor = aPlayer.getEquipmentInSlot(i);
			if (GT_ModHandler.isElectricItem(tArmor)) {
				IElectricItem tArmorItem = (IElectricItem)tArmor.getItem();
				if (tArmorItem.canProvideEnergy(tArmor) && tArmorItem.getTier(tArmor) >= getTier(aStack)) {
					double tCharge = ElectricItem.manager.discharge(tArmor, charge(aStack, Integer.MAX_VALUE-1, Integer.MAX_VALUE, T, T), Integer.MAX_VALUE, T, T, F);
					if (tCharge > 0) {
						charge(aStack, tCharge, Integer.MAX_VALUE, T, F);
						if (aPlayer instanceof EntityPlayer) {
							Container tContainer = ((EntityPlayer)aPlayer).openContainer;
							if (tContainer != null) tContainer.detectAndSendChanges();
						}
					}
				}
			}
		}
	}
	
	public long getRealCharge(ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergyStored(TD.Energy.EU, aStack);
	}
	
	public boolean setCharge(ItemStack aStack, long aCharge) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return F;
		tStats.setEnergyStored(TD.Energy.EU, aStack, aCharge);
		return T;
	}
	
	public short getChargedMetaData(ItemStack aStack) {return UT.Stacks.meta(aStack);}
	public short getEmptyMetaData(ItemStack aStack) {return UT.Stacks.meta(aStack);}
	
	public FluidStack getFluid(ItemStack aStack) {
		return getFluidContent(aStack);
	}
	
	public int getCapacity(ItemStack aStack) {
		Long[] tStats = getFluidContainerStats(aStack);
		return tStats==null?0:(int)Math.max(0, tStats[0]);
	}
	
	public int fill(ItemStack aStack, FluidStack aFluid, boolean doFill) {
		if (aStack == null || aStack.stackSize != 1) return 0;
		
		ItemStack tStack = UT.Fluids.fillFluidContainer(aFluid, aStack, F, F, F);
		if (tStack != null) {
			aStack.setItemDamage(UT.Stacks.meta(tStack));
			aStack.func_150996_a(tStack.getItem());
			return UT.Fluids.getFluidForFilledItem(tStack, F).amount;
		}
		
		Long[] tStats = getFluidContainerStats(aStack);
		if (tStats == null || tStats[0] <= 0 || aFluid == null || aFluid.amount <= 0 || !isAllowedToFill(aStack, aFluid)) return 0;
		
		FluidStack tFluid = getFluidContent(aStack);
		
		if (tFluid == null) {
			if (aFluid.amount <= tStats[0]) {
				if (doFill) {
					setFluidContent(aStack, aFluid);
				}
				return aFluid.amount;
			}
			if (doFill) {
				tFluid = aFluid.copy();
				tFluid.amount = (int)(long)tStats[0];
				setFluidContent(aStack, tFluid);
			}
			return (int)(long)tStats[0];
		}
		
		if (!tFluid.isFluidEqual(aFluid)) return 0;
		
		int space = (int)(long)tStats[0] - tFluid.amount;
		if (aFluid.amount <= space) {
			if (doFill) {
				tFluid.amount += aFluid.amount;
				setFluidContent(aStack, tFluid);
			}
			return aFluid.amount;
		}
		if (doFill) {
			tFluid.amount = (int)(long)tStats[0];
			setFluidContent(aStack, tFluid);
		}
		return space;
	}
	
	public boolean isAllowedToFill(ItemStack aStack, FluidStack aFluid) {
		return T;
	}
	
	public FluidStack drain(ItemStack aStack, int aMaxDrain, boolean aDoDrain) {
		if (aStack == null || aStack.stackSize != 1) return null;
		
		FluidStack tFluid = UT.Fluids.getFluidForFilledItem(aStack, F);
		if (tFluid != null && aMaxDrain >= tFluid.amount) {
			if (aDoDrain) {
				ItemStack tStack = UT.Stacks.container(aStack, F);
				if (tStack == null) {
					aStack.stackSize = 0;
					return tFluid;
				}
				aStack.setItemDamage(UT.Stacks.meta(tStack));
				aStack.func_150996_a(tStack.getItem());
			}
			return tFluid;
		}
		
		Long[] tStats = getFluidContainerStats(aStack);
		if (tStats == null || tStats[0] <= 0) return null;
		
		tFluid = getFluidContent(aStack);
		if (tFluid == null) return null;
		
		if (tFluid.amount < aMaxDrain) aMaxDrain = tFluid.amount;
		if (aDoDrain) {
			tFluid.amount -= aMaxDrain;
			setFluidContent(aStack, tFluid);
		}
		
		FluidStack tDrained = tFluid.copy();
		tDrained.amount = aMaxDrain;
		return tDrained;
	}
	
	public FluidStack getFluidContent(ItemStack aStack) {
		Long[] tStats = getFluidContainerStats(aStack);
		if (tStats == null || tStats[0] <= 0) return UT.Fluids.getFluidForFilledItem(aStack, F);
		NBTTagCompound tNBT = aStack.getTagCompound();
		return tNBT==null?null:FluidStack.loadFluidStackFromNBT(tNBT.getCompoundTag("gt.fluidcontent"));
	}
	
	public void setFluidContent(ItemStack aStack, FluidStack aFluid) {
		NBTTagCompound tNBT = aStack.getTagCompound();
		if (tNBT == null) tNBT = new NBTTagCompound(); else tNBT.removeTag("gt.fluidcontent");
		if (aFluid != null && aFluid.amount > 0) tNBT.setTag("gt.fluidcontent", aFluid.writeToNBT(new NBTTagCompound()));
		aStack.setTagCompound(tNBT.hasNoTags()?null:tNBT);
		isItemStackUsable(aStack);
	}
	
	@Override
	public int getItemStackLimit(ItemStack aStack) {
		if (aStack.hasTagCompound() && aStack.getTagCompound().hasKey(NBT_ENERGY)) return 1;
		Long[] tStats = getFluidContainerStats(aStack);
		if (tStats != null) return (int)(long)tStats[1];
		return UT.Code.bindStack(getDefaultStackLimit(aStack));
	}
	
	public int getDefaultStackLimit(ItemStack aStack) {
		return 64;
	}
	
	@Optional.Method(modid = ModIDs.IC2)
	@Override public IElectricItemManager getManager(ItemStack aStack) {return this;} // We are our own Manager
	@Override public Item getChargedItem(ItemStack itemStack) {return this;}
	@Override public Item getEmptyItem(ItemStack itemStack) {return this;}
    @Override public int getTier(ItemStack aStack) {IItemEnergy tStats = getEnergyStats(aStack); return tStats==null?Integer.MAX_VALUE:UT.Code.tierMax(tStats.getEnergySizeInputMax(TD.Energy.EU, aStack));}
	@Override public String getToolTip(ItemStack aStack) {return null;} // This has its own ToolTip Handler, no need to let the IC2 Handler screw us up at this Point
	@Override public final boolean getShareTag() {return T;} // just to be sure.
	@Override public int getItemEnchantability() {return 0;}
	@Override public boolean isBookEnchantable(ItemStack aStack, ItemStack aBook) {return F;}
	@Override public boolean getIsRepairable(ItemStack aStack, ItemStack aMaterial) {return F;}
	
	@Override
	public boolean isEnergyType(TagData aEnergyType, ItemStack aStack, boolean aEmitting) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return F;
		return tStats.isEnergyType(aEnergyType, aStack, aEmitting);
	}
	
	@Override
	public long doEnergyInjection(TagData aEnergyType, ItemStack aStack, long aSize, long aAmount, IInventory aInventory, World aWorld, int aX, int aY, int aZ, boolean aDoInject) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.doEnergyInjection(aEnergyType, aStack, aSize, aAmount, aInventory, aWorld, aX, aY, aZ, aDoInject);
	}
	
	@Override
	public long doEnergyExtraction(TagData aEnergyType, ItemStack aStack, long aSize, long aAmount, IInventory aInventory, World aWorld, int aX, int aY, int aZ, boolean aDoExtract) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.doEnergyExtraction(aEnergyType, aStack, aSize, aAmount, aInventory, aWorld, aX, aY, aZ, aDoExtract);
	}
	
	@Override
	public boolean useEnergy(TagData aEnergyType, ItemStack aStack, long aEnergyAmount, EntityLivingBase aPlayer, IInventory aInventory, World aWorld, int aX, int aY, int aZ, boolean aDoUse) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return F;
		return tStats.useEnergy(aEnergyType, aStack, aEnergyAmount, aPlayer, aInventory, aWorld, aX, aY, aZ, aDoUse);
	}
	
	@Override
	public ItemStack setEnergyStored(TagData aEnergyType, ItemStack aStack, long aAmount) {
		if (aEnergyType == TD.Energy.EU) setCharge(aStack, aAmount);
		return aStack;
	}
	
	@Override
	public long getEnergyStored(TagData aEnergyType, ItemStack aStack) {
		return aEnergyType == TD.Energy.EU ? getRealCharge(aStack) : 0;
	}
	
	@Override
	public long getEnergyCapacity(TagData aEnergyType, ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergyCapacity(aEnergyType, aStack);
	}
	
	@Override
	public long getEnergySizeInputMin(TagData aEnergyType, ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergySizeInputMin(aEnergyType, aStack);
	}
	
	@Override
	public long getEnergySizeOutputMin(TagData aEnergyType, ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergySizeOutputMin(aEnergyType, aStack);
	}
	
	@Override
	public long getEnergySizeInputRecommended(TagData aEnergyType, ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergySizeInputRecommended(aEnergyType, aStack);
	}
	
	@Override
	public long getEnergySizeOutputRecommended(TagData aEnergyType, ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergySizeOutputRecommended(aEnergyType, aStack);
	}
	
	@Override
	public long getEnergySizeInputMax(TagData aEnergyType, ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergySizeInputMax(aEnergyType, aStack);
	}
	
	@Override
	public long getEnergySizeOutputMax(TagData aEnergyType, ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return 0;
		return tStats.getEnergySizeOutputMax(aEnergyType, aStack);
	}
	
	@Override
	public Collection<TagData> getEnergyTypes(ItemStack aStack) {
		IItemEnergy tStats = getEnergyStats(aStack);
		if (tStats == null) return Collections.EMPTY_LIST;
		return tStats.getEnergyTypes(aStack);
	}
	
	@Override
	public boolean canEnergyInjection(TagData aEnergyType, ItemStack aStack, long aSize) {
		IItemEnergy tStats = getEnergyStats(aStack);
		return tStats != null && tStats.canEnergyInjection(aEnergyType, aStack, aSize);
	}
	
	@Override
	public boolean canEnergyExtraction(TagData aEnergyType, ItemStack aStack, long aSize) {
		IItemEnergy tStats = getEnergyStats(aStack);
		return tStats != null && tStats.canEnergyExtraction(aEnergyType, aStack, aSize);
	}
	
	public abstract IItemEnergy getEnergyStats(ItemStack aStack);
	public abstract Long[] getFluidContainerStats(ItemStack aStack);
}