package gregapi.tileentity.energy;

import static gregapi.data.CS.*;
import gregapi.code.TagData;
import gregapi.data.LH;
import gregapi.data.LH.Chat;
import gregapi.data.TD;
import gregapi.tileentity.machines.ITileEntitySwitchableOnOff;
import gregapi.util.UT;

import java.util.List;

import net.minecraft.entity.Entity;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;

/**
 * @author Gregorius Techneticies
 */
public abstract class TileEntityBase11Bidirectional extends TileEntityBase10EnergyConverter implements ITileEntitySwitchableOnOff {
	protected boolean mReversed = F;
	
    @Override
	public void readFromNBT2(NBTTagCompound aNBT) {
    	super.readFromNBT2(aNBT);
        if (aNBT.hasKey(NBT_REVERSED)) mReversed = aNBT.getBoolean(NBT_REVERSED);
    }
    
    @Override
	public void writeToNBT2(NBTTagCompound aNBT) {
    	super.writeToNBT2(aNBT);
        UT.NBT.setBoolean(aNBT, NBT_REVERSED, mReversed);
    }
    
	@Override public boolean isEnergyType					(TagData aEnergyType, byte aSide, boolean aEmitting) {return aEmitting != mReversed ? aEnergyType == mEnergyTypeEmitted : aEnergyType == mEnergyTypeAccepted;}
	@Override public long getEnergySizeOutputMin			(TagData aEnergyType, byte aSide) {return (mReversed?mInput:mOutput)<=8?1:(mReversed?mInput:mOutput)/2;}
	@Override public long getEnergySizeOutputRecommended	(TagData aEnergyType, byte aSide) {return  mReversed?mInput:mOutput;}
	@Override public long getEnergySizeOutputMax			(TagData aEnergyType, byte aSide) {return (mReversed?mInput:mOutput)*2;}
	@Override public long getEnergySizeInputMin				(TagData aEnergyType, byte aSide) {return (mReversed?mOutput:mInput)<=8?1:(mReversed?mOutput:mInput)/2;}
	@Override public long getEnergySizeInputRecommended		(TagData aEnergyType, byte aSide) {return  mReversed?mOutput:mInput;}
	@Override public long getEnergySizeInputMax				(TagData aEnergyType, byte aSide) {return (mReversed?mOutput:mInput)*2;}
	@Override public long getMaxBuffered					(TagData aEnergyType, byte aSide) {return Math.max(mInput, mOutput*mMultiplier);}
	@Override public boolean isInput (byte aSide) {return mReversed ? aSide != mFacing : aSide == mFacing;}
	@Override public boolean isOutput(byte aSide) {return mReversed ? aSide == mFacing : aSide != mFacing;}
	@Override public String getLocalisedInputSide () {return LH.get(LH.FACE_FRONT);}
	@Override public String getLocalisedOutputSide() {return LH.get(LH.FACE_ANYBUT_FRONT);}
	
	@Override
	public void addToolTips(List aList, ItemStack aStack, boolean aF3_H) {
		super.addToolTips(aList, aStack, aF3_H);
		aList.add(Chat.DGRAY	+ LH.get(LH.TOOL_TO_SET_DIRECTION_MONKEY_WRENCH));
	}
	
	@Override
	public long onToolClick2(String aTool, long aRemainingDurability, long aQuality, Entity aPlayer, List<String> aChatReturn, IInventory aPlayerInventory, boolean aSneaking, ItemStack aStack, byte aSide, float aHitX, float aHitY, float aHitZ) {
		long rReturn = super.onToolClick2(aTool, aRemainingDurability, aQuality, aPlayer, aChatReturn, aPlayerInventory, aSneaking, aStack, aSide, aHitX, aHitY, aHitZ);
		if (rReturn > 0) return rReturn;
		
		if (isClientSide()) return 0;
		if (aTool.equals(TOOL_monkeywrench)) {mReversed=!mReversed; causeBlockUpdate(); doEnetUpdate(); return 10000;}
		
		return 0;
	}
	
	@Override
	public void doConversion(long aTimer) {
		long tOutput = UT.Code.units(mEnergy, mReversed?mOutput*mMultiplier:mInput, mReversed?mInput:mOutput, F), tMaxOut = getEnergySizeOutputMax(mReversed?mEnergyTypeAccepted:mEnergyTypeEmitted, SIDE_ANY);
		mActive = (tOutput >= getEnergySizeOutputMin(mReversed?mEnergyTypeAccepted:mEnergyTypeEmitted, SIDE_ANY));
		mEmitsEnergy = F;
		
		if (mActive) {
			if (tOutput > tMaxOut && TD.Energy.ALL_COMSUMPTION_LIMITED.contains(mReversed?mEnergyTypeEmitted:mEnergyTypeAccepted)) tOutput = tMaxOut;
			if (tOutput > tMaxOut) {
				overcharge(tOutput, mReversed?mEnergyTypeAccepted:mEnergyTypeEmitted);
			} else {
				if (TD.Energy.ALL_SIZE_IRRELEVANT.contains(mReversed?mEnergyTypeAccepted:mEnergyTypeEmitted)) {
					long tEmittedPackets = ITileEntityEnergy.Util.emitEnergyToNetwork(mReversed?mEnergyTypeAccepted:mEnergyTypeEmitted, 1, tOutput*(mReversed?1:mMultiplier), this);
					if (!mWasteEnergy) mEnergy -= UT.Code.units(tEmittedPackets, mReversed?mInput:mOutput*mMultiplier, mReversed?mOutput*mMultiplier:mInput, T);
					mEmitsEnergy = (tEmittedPackets > 0);
				} else {
					long tEmittedPackets = ITileEntityEnergy.Util.emitEnergyToNetwork(mReversed?mEnergyTypeAccepted:mEnergyTypeEmitted, tOutput, mReversed?1:mMultiplier, this);
					if (!mWasteEnergy) mEnergy -= UT.Code.units(tEmittedPackets * tOutput, mReversed?mInput:mOutput*mMultiplier, mReversed?mOutput*mMultiplier:mInput, T);
					mEmitsEnergy = (tEmittedPackets > 0);
				}
			}
		}
		if (mWasteEnergy) mEnergy = Math.max(0, mEnergy-getEnergySizeInputMax(mEnergyTypeAccepted, SIDE_ANY));
	}
}